from pathlib import Path
import click
import webbrowser

import geoclustering.clustering as clustering
import geoclustering.encoding as encoding
import geoclustering.io as io


@click.command(
    help="Tool to cluster geolocations. A cluster is created when a certain number of points (defined with --size) each are within a given distance (defined with --distance) of at least one other point in the cluster. Input is supplied as a csv file. At a minimum, each row needs to have a 'lat' and a 'lon' column. Other rows are reflected to the output."
)
@click.option(
    "--distance",
    "-d",
    type=click.FLOAT,
    required=True,
    help="(in km) Max. distance between two points in a cluster.",
)
@click.option(
    "--size",
    "-s",
    type=click.INT,
    required=True,
    help="Min. number of points in a cluster.",
)
@click.option(
    "--output",
    "-o",
    type=click.Path(exists=False),
    default="output",
    help="Output directory for results. Default: ./output",
)
@click.option(
    "--algorithm",
    "-a",
    type=click.Choice(
        ["dbscan", "optics"],
        case_sensitive=False,
    ),
    default="dbscan",
    help="Clustering algorithm to be used. `optics` produces tighter clusters but is slower. Default: dbscan",
)
@click.option(
    "--open",
    "_open",
    is_flag=True,
    help="Open the generated visualization in the default browser automatically.",
)
@click.option("--debug", is_flag=True, help="Print debug output.")
@click.argument("filename", type=click.Path(exists=True))
def main(distance, size, output, filename, algorithm, _open, debug):
    def print_debug(s):
        if debug:
            click.secho(s, fg="bright_black")

    df = io.read_csv_file(filename)
    print_debug(f"Read {len(df)} valid coordinates from {Path(filename).absolute()}")

    clusters = clustering.cluster_locations(
        df=df, algorithm=algorithm, radius_km=distance, min_cluster_size=size
    )

    if not bool(clusters):
        click.secho("Did not find clusters matching input parameters.", fg="yellow")
        return

    print_debug(f"Found {len(clusters)} valid clusters using {algorithm}")

    encoded = encoding.encode_clusters(clusters)
    io.write_output_file(output, "result.txt", encoded["string"])
    io.write_output_file(output, "result.json", encoded["json"])
    io.write_output_file(output, "result.geojson", encoded["geojson"])
    io.write_output_file(output, "result.csv", encoded["csv"])

    vis = io.write_visualization(output, "result.html", encoded["geojson"])
    if vis is None:
        print_debug("Skipped generating visualization: kepler is not installed.")

    click.echo(f"Output files saved to {Path(output).absolute()}")

    if _open:
        if vis:
            webbrowser.open_new_tab("file://" + str(vis.absolute()))
            print_debug("Opened visualization in default browser.")
        else:
            click.secho(
                "Can't open kepler.gl: package not installed. Please re-install geoclustering with `pip install geoclustering[full]`.",
                fg="yellow",
            )

    click.secho("Clustering completed.", fg="green")


if __name__ == "__main__":
    main()



import json
import numpy as np
import geojson
import csv
import io  # not io.py


class NpEncoder(json.JSONEncoder):
    """JSONEncoder with support for numpy's numerical types."""

    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        return super(NpEncoder, self).default(obj)


class StringEncoder:
    """Encodes clustering result as a string."""

    def __init__(self):
        self.state = []

    def visitor(self, cluster_id, cluster):
        self.state.append("Cluster {}".format(cluster_id))

        for record in cluster:
            s = []
            for key, val in record.items():
                s.append("{} {}".format(key, val))
            self.state.append(", ".join(s))

        # separate clusters by an empty line.
        self.state.append("")

    def get(self):
        return "\n".join(self.state)


class JSONEncoder:
    """Encodes clustering result as a JSON array."""

    def __init__(self):
        self.state = []

    def visitor(self, cluster_id, cluster):
        cluster_data = {"cluster_id": cluster_id, "points": []}

        for record in cluster:
            cluster_data["points"].append(record)
        self.state.append(cluster_data)

    def get(self):
        return json.dumps(self.state, cls=NpEncoder)


class GeoJSONEncoder:
    def __init__(self):
        self.state = []

    def visitor(self, cluster_id, cluster):
        for record in cluster:
            props = {
                **record,
                "cluster_id": cluster_id,
            }

            lon = float(props.pop("lon"))
            lat = float(props.pop("lat"))

            point = geojson.Point((lon, lat))
            self.state.append(geojson.Feature(geometry=point, properties=props))

    def get(self):
        return json.dumps(geojson.FeatureCollection(self.state), cls=NpEncoder)


class CSVEncoder:
    """Encodes clustering result as a CSV"""

    def __init__(self):
        self.state = io.StringIO()
        self.writer = False

    def visitor(self, cluster_id, cluster):
        if not self.writer:
            self.writer = csv.DictWriter(
                self.state,
                fieldnames=["cluster_id"] + list(cluster[0].keys()),
                quoting=csv.QUOTE_NONNUMERIC,
                lineterminator="\n",
            )
            self.writer.writeheader()

        for record in cluster:
            self.writer.writerow({**record, "cluster_id": cluster_id})

    def get(self):
        return self.state.getvalue()


def encode_clusters(clusters):
    json_encoder = JSONEncoder()
    geojson_encoder = GeoJSONEncoder()
    string_encoder = StringEncoder()
    csv_encoder = CSVEncoder()

    encoders = [json_encoder, geojson_encoder, string_encoder, csv_encoder]
    for cluster_id, cluster in clusters.items():
        for encoder in encoders:
            encoder.visitor(cluster_id, cluster)

    return {
        "json": json_encoder.get(),
        "geojson": geojson_encoder.get(),
        "string": string_encoder.get(),
        "csv": csv_encoder.get(),
    }



from setuptools import setup

# version.py defines the VERSION and VERSION_SHORT variables.
# We use exec here so we don't import cached_path whilst setting up.
VERSION = {}  # type: ignore
with open("geoclustering/version.py", "r") as version_file:
    exec(version_file.read(), VERSION)

setup(
    name="geoclustering",
    version=VERSION["VERSION"],
    description="📍 command-line tool for clustering geolocations.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    author="Bellingcat",
    author_email="tech@bellingcat.com",
    license="MIT",
    packages=["geoclustering"],
    package_data={"geoclustering": ["kepler_config.json"]},
    keywords=["cluster", "gis", "pattern-analysis"],
    entry_points={"console_scripts": ["geoclustering = geoclustering.__main__:main"]},
    install_requires=[
        "click",
        "geojson",
        "numpy",
        "pandas",
        "scikit-learn",
    ],
    extras_require={
        "dev": ["black", "wheel", "pre-commit", "pytest"],
        "full": ["keplergl"],
    },
    include_package_data=True,
    zip_safe=False,
)



from pathlib import Path
from pkg_resources import resource_filename
import json
import pandas as pd
import numpy as np
import os
import sys

# kepler is optional, check if installed.
try:
    from keplergl import KeplerGl
except:
    has_kepler = False
else:
    has_kepler = True


class HiddenPrints:
    """Disables stdout prints for a block of code."""

    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


def is_valid_lat(val: str) -> bool:
    """Given a string, check if it corresponds to a valid decimal latitude value"""
    try:
        val = float(val)
        return val >= -90 and val <= 90
    except:
        return False


def is_valid_lon(val: str) -> bool:
    """Given a string, check if it corresponds to a valid decimal longitude value"""
    try:
        val = float(val)
        return val >= -180 and val <= 180
    except:
        return False


def is_not_none(val: any) -> bool:
    return val is not None


def read_csv_file(filename):
    """Read input csv file, dropping rows that don't have valid location data."""
    # replace NaN for all fields not to break kepler parsing.
    df = pd.read_csv(filename).replace({np.nan: None})

    # construct an index of values with valid lat & lon.
    valid_index = df.lat.apply(is_valid_lat) & df.lon.apply(is_valid_lon)
    df_invalid = df[~valid_index]

    count_invalid = len(df_invalid)
    if count_invalid:
        df_not_empty = df_invalid[
            (df_invalid.lat.apply(is_not_none) | df_invalid.lon.apply(is_not_none))
        ]

        count_not_empty = len(df_not_empty)
        count_empty = count_invalid - count_not_empty

        if count_empty:
            print(f"Removed {count_empty} empty coordinate pairs.")

        if count_not_empty:
            print(f"Removed {count_not_empty} invalid coordinate pairs:")
            print(df_not_empty[["lat", "lon"]].to_string())

    return df[valid_index]


def ensure_file_path(dirname, filename):
    """Ensure a parent directory exists for a file."""
    path = Path(dirname)
    path.mkdir(parents=True, exist_ok=True)
    return path / filename


def write_output_file(dirname, filename, data):
    """Write a file, ensuring parent directories."""
    filepath = ensure_file_path(dirname, filename)

    with open(filepath, "w") as f:
        f.write(data)

    return filepath


def write_visualization(dirname, filename, data):
    """Write a visualization, ensuring parent directories."""

    if not has_kepler:
        return None

    # Hide kepler stdout output.
    with HiddenPrints():
        map = KeplerGl()

    map.add_data(data=data, name="clusters")

    # config configures a default color scheme for our clusters layer.
    config_file = resource_filename("geoclustering", "kepler_config.json")
    with open(config_file) as f:
        map.config = json.loads(f.read())

    filepath = ensure_file_path(dirname, filename)

    # Hide kepler stdout output.
    with HiddenPrints():
        map.save_to_html(file_name=str(filepath), center_map=True)

    return filepath
